__all__ = ["optimize"]

import pymc as pm
from pytensor.graph.basic import graph_inputs
from pytensor.tensor.variable import TensorConstant, TensorVariable


def optimize(start=None, vars=None, **kwargs):
    """A thin wrapper around ``pymc.find_MAP`` that actually does something sensible :D"""
    if vars is not None:
        if not isinstance(vars, (list, tuple)):
            vars = [vars]

        # In PyMC >= 5, model context is required to replace rvs with values
        # https://github.com/pymc-devs/pymc/pull/6281
        # https://www.pymc.io/projects/docs/en/stable/learn/core_notebooks/pymc_pytensor.html#pymc
        model = kwargs.get("model")
        model = pm.modelcontext(model)

        # find_MAP only supports passing in members of free_RVs, so let's deal
        # with that here...
        vars = model.replace_rvs_by_values(vars)
        vars = [
            v
            for v in graph_inputs(vars)
            if isinstance(v, TensorVariable)
            and not isinstance(v, TensorConstant)
        ]

    return pm.find_MAP(start=start, vars=vars, **kwargs)



# -*- coding: utf-8 -*-

import logging

import numpy as np
import pymc as pm
import pytest
from scipy.stats import kstest

from pymc_ext.distributions import angle, unit_disk


class _Base:
    random_seed = 20160911

    def _sample(self, **kwargs):
        logger = logging.getLogger("pymc3")
        logger.propagate = False
        logger.setLevel(logging.ERROR)
        kwargs["draws"] = kwargs.get("draws", 1000)
        kwargs["progressbar"] = kwargs.get("progressbar", False)
        kwargs["return_inferencedata"] = kwargs.get(
            "return_inferencedata", False
        )
        return pm.sample(**kwargs)

    def _model(self, **kwargs):
        np.random.seed(self.random_seed)
        return pm.Model(**kwargs)


class TestBase(_Base):
    def test_unit_disk(self):
        with self._model():
            unit_disk("x", "y", shape=(3,), initval=0.01 * np.ones((2, 3)))
            trace = self._sample()

        theta = np.arctan2(trace["x"], trace["y"])
        radius = trace["x"] ** 2 + trace["y"] ** 2

        # Make sure that the unit constraint is satisfied
        assert np.all(radius <= 1.0)

        # The angle should be uniformly distributed
        cdf = lambda x: np.clip((x + np.pi) / (2 * np.pi), 0, 1)  # NOQA
        for i in range(theta.shape[1]):
            s, _ = kstest(theta[:, i], cdf)
            assert s < 0.05

        # As should the radius component
        cdf = lambda x: np.clip(x, 0, 1)  # NOQA
        for i in range(radius.shape[1]):
            s, p = kstest(radius[:, i], cdf)
            assert s < 0.05

    @pytest.mark.parametrize("regularization", [None, 10.0])
    def test_angle(self, regularization):
        with self._model():
            angle("theta", shape=(5, 2), regularization=regularization)
            trace = self._sample(draws=2000)

        # The angle should be uniformly distributed
        theta = trace["theta"]
        theta = np.reshape(theta, (len(theta), -1))
        cdf = lambda x: np.clip((x + np.pi) / (2 * np.pi), 0, 1)  # NOQA
        for i in range(theta.shape[1]):
            s, _ = kstest(theta[:, i], cdf)
            assert s < 0.05



__all__ = ["angle", "unit_disk"]

import warnings

import numpy as np
import pymc as pm
import pytensor.tensor as pt


def angle(name, *, regularization=10.0, **kwargs):
    """An angle constrained to be in the range -pi to pi

    The actual sampling is performed in the two dimensional vector space
    proportional to ``(sin(theta), cos(theta))`` so that the sampler doesn't see
    a discontinuity at pi.

    The regularization parameter can be used to improve sampling performance
    when the value of the angle is well constrained. It removes prior mass near
    the origin in the sampling space, which can lead to bad geometry when the
    angle is poorly constrained, but better performance when it is. The default
    value of ``10.0`` is a good starting point.
    """
    shape = kwargs.get("shape", ())
    initval = kwargs.pop("initval", pt.broadcast_to(0.0, shape))
    x1 = pm.Normal(f"__{name}_angle1", initval=np.sin(initval), **kwargs)
    x2 = pm.Normal(f"__{name}_angle2", initval=np.cos(initval), **kwargs)
    if regularization is not None:
        pm.Potential(
            f"__{name}_regularization",
            regularization * pt.log(x1**2 + x2**2),
        )
    return pm.Deterministic(name, pt.arctan2(x1, x2))


def unit_disk(name_x, name_y, **kwargs):
    """Two dimensional parameters constrained to live within the unit disk

    This returns two distributions whose sum of squares will be in the range
    ``[0, 1)``. For example, in this code block:

    .. code-block:: python

        x, y = unit_disk("x", "y") radius_sq = x**2 + y**2

    the tensor ``radius_sq`` will always have a value in the range ``[0, 1)``.

    Args:
        name_x: The name of the first distribution.
        name_y: The name of the second distribution.
    """
    initval = kwargs.pop("initval", [0.0, 0.0])
    kwargs["lower"] = -1.0
    kwargs["upper"] = 1.0
    x1 = pm.Uniform(name_x, initval=initval[0], **kwargs)
    x2 = pm.Uniform(
        f"__{name_y}_unit_disk",
        initval=initval[1] * np.sqrt(1 - initval[0] ** 2),
        **kwargs,
    )
    norm = pt.sqrt(1 - x1**2)
    pm.Potential(f"__{name_y}_jacobian", pt.log(norm))
    return x1, pm.Deterministic(name_y, x2 * norm)



__all__ = [
    "angle",
    "unit_disk",
    "eval_in_model",
    "optimize",
    "sample_inference_data",
]

from pymc_ext.distributions import angle, unit_disk
from pymc_ext.optim import optimize
from pymc_ext.pymc_ext_version import __version__ as __version__
from pymc_ext.utils import eval_in_model, sample_inference_data

__uri__ = "https://github.com/exoplanet-dev/pymc-ext"
__author__ = "Daniel Foreman-Mackey"
__email__ = "foreman.mackey@gmail.com"
__license__ = "MIT"
__description__ = 'PyMC Extras extracted from the "exoplanet" library'



import numpy as np
import pymc as pm
import pytensor.tensor as pt
import pytest

from pymc_ext.optim import optimize


def test_optimize(seed=1234):
    random = np.random.default_rng(seed)
    x_val = random.standard_normal(size=(5, 3))
    with pm.Model():
        x = pm.Normal("x", shape=x_val.shape, initval=x_val)
        y = pm.Normal("y", shape=x_val.shape, initval=x_val)
        soln1 = optimize()
        soln2 = optimize(vars=[x + y])

    assert np.allclose(soln1["x"], 0.0)
    assert np.allclose(soln1["y"], 0.0)
    assert np.allclose(soln2["x"], 0.0)
    assert np.allclose(soln2["y"], 0.0)



import numpy as np
import pymc as pm
import pytest
from scipy.stats import invgamma

from pymc_ext.utils import estimate_inverse_gamma_parameters, eval_in_model


def test_eval_in_model(seed=123409):
    np.random.seed(seed)
    x_val = np.random.randn(5, 3)
    x_val2 = np.random.randn(5, 3)
    with pm.Model():
        x = pm.Normal("x", shape=x_val.shape, initval=x_val)
        assert np.allclose(eval_in_model(x), x_val)
        assert np.allclose(eval_in_model(x, {"x": x_val2}), x_val2)


def test_eval_in_model_uniform(seed=123409):
    # test_eval_in_model has unconstrained (-inf, inf) variables only
    # Uniform has implicit transform in PyMC so check that this works too with
    # eval_in_model
    rng = np.random.default_rng(seed)
    x_val = rng.uniform(size=(5, 3))
    with pm.Model():
        x = pm.Uniform("x", shape=x_val.shape, initval=x_val)

        assert np.allclose(eval_in_model(x), x_val)


def test_eval_in_model_list(seed=123409):
    # The utils.Evaluator class handles list of variables differently
    # from single variables, so we test this here.
    rng = np.random.default_rng(seed)
    x_val = rng.uniform(size=(5, 3))
    y_val = rng.standard_normal()

    with pm.Model():
        x = pm.Uniform("x", shape=x_val.shape, initval=x_val)
        y = pm.Normal("y", initval=y_val)
        x_eval, y_eval = eval_in_model([x, y])
        assert np.allclose(x_eval, x_val)
        assert np.allclose(y_eval, y_val)


@pytest.mark.parametrize(
    "lower, upper, target",
    [(1.0, 2.0, 0.01), (0.01, 0.1, 0.1), (10.0, 25.0, 0.01)],
)
def test_estimate_inverse_gamma_parameters(lower, upper, target):
    np.random.seed(20200409)

    params = estimate_inverse_gamma_parameters(lower, upper, target=target)
    dist = invgamma(params["alpha"], scale=params["beta"])
    assert np.allclose(dist.cdf(lower), target)
    assert np.allclose(1 - dist.cdf(upper), target)

    samples = pm.draw(pm.InverseGamma.dist(**params), draws=10000)
    assert np.allclose(
        (samples < lower).sum() / len(samples), target, atol=1e-2
    )
    assert np.allclose(
        (samples > upper).sum() / len(samples), target, atol=1e-2
    )



__all__ = [
    "Evaluator",
    "eval_in_model",
    "sample_inference_data",
    "estimate_inverse_gamma_parameters",
]

import numpy as np
import pymc as pm
from scipy.optimize import root
from scipy.special import gammaincc


class Evaluator:
    """Class to compile and evaluate components of a PyMC model

    Args:
        outs: The random variable, tensor, or list thereof to evaluate
        model (Optional): PyMC model in which the variable are defined.
                            Tries to infer current model context if None.
        **kwargs: All other kwargs are passed to pymc.pytensorf.compile_pymc.
    """

    def __init__(self, outs, model=None, **kwargs):
        # In PyMC >= 5, model context is required to replace rvs with values
        # https://github.com/pymc-devs/pymc/pull/6281
        # https://www.pymc.io/projects/docs/en/stable/learn/core_notebooks/pymc_pytensor.html#pymc
        model = pm.modelcontext(model)
        if isinstance(outs, (tuple, list)):
            self.out_values = model.replace_rvs_by_values(outs)
        else:
            self.out_values = model.replace_rvs_by_values([outs])[0]
        self.in_values = pm.inputvars(self.out_values)
        self.func = pm.pytensorf.compile_pymc(
            self.in_values, self.out_values, **kwargs
        )

    def __call__(self, point):
        args = [point[x.name] for x in self.in_values]
        return self.func(*args)


def eval_in_model(outs, point=None, model=None, seed=None, **kwargs):
    """Evaluate a PyTensor tensor or PyMC variable in a PyMC model

    This method builds a Theano function for evaluating a node in the graph
    given the required parameters. This will also cache the compiled Theano
    function in the current ``pymc3.Model`` to reduce the overhead of calling
    this function many times.

    Args:
        outs: The variable, tensor, or list thereof to evaluate.
        point (Optional): A ``dict`` of input parameter values. This can be
            ``model.initial_point`` (default), the result of ``pymc.find_MAP``,
            a point in a ``pymc3.MultiTrace`` or any other representation of
            the input parameters.
    """
    if point is None:
        model = pm.modelcontext(model)
        point = model.initial_point(random_seed=seed)
    return Evaluator(outs, model=model, **kwargs)(point)


def sample_inference_data(idata, size=1, random_seed=None, group="posterior"):
    """Generate random samples from an InferenceData object

    Args:
        idata: The ``InferenceData``.
        size: The number of samples to generate.
        random_seed: The seed for the random number generator.
        group: The ``InferenceData`` group to sample from.
    """
    random = np.random.default_rng(random_seed)
    idata = idata[group].stack(sample=("chain", "draw"))
    num_sample = idata.dims["sample"]

    for i in random.integers(num_sample, size=size):
        yield {k: v.values for k, v in idata.isel({"sample": i}).items()}


def sample(**kwargs):
    kwargs["target_accept"] = kwargs.get("target_accept", 0.9)
    kwargs["init"] = kwargs.get("init", "adapt_full")
    return pm.sample(**kwargs)


def estimate_inverse_gamma_parameters(
    lower, upper, target=0.01, initial=None, **kwargs
):
    r"""Estimate an inverse Gamma with desired tail probabilities
    This method numerically solves for the parameters of an inverse Gamma
    distribution where the tails have a given probability. In other words
    :math:`P(x < \mathrm{lower}) = \mathrm{target}` and similarly for the
    upper bound. More information can be found in `part 4 of this blog post
    <https://betanalpha.github.io/assets/case_studies/gp_part3/part3.html>`_.
    Args:
        lower (float): The location of the lower tail
        upper (float): The location of the upper tail
        target (float, optional): The desired tail probability
        initial (ndarray, optional): An initial guess for the parameters
            ``alpha`` and ``beta``
    Raises:
        RuntimeError: If the solver does not converge.
    Returns:
        dict: A dictionary with the keys ``alpha`` and ``beta`` for the
        parameters of the distribution.
    """
    lower, upper = np.sort([lower, upper])
    if initial is None:
        initial = np.array([2.0, 0.5 * (lower + upper)])
    if np.shape(initial) != (2,) or np.any(np.asarray(initial) <= 0.0):
        raise ValueError("invalid initial guess")

    def obj(x):
        a, b = np.exp(x)
        return np.array(
            [
                gammaincc(a, b / lower) - target,
                1 - gammaincc(a, b / upper) - target,
            ]
        )

    result = root(obj, np.log(initial), method="hybr", **kwargs)
    if not result.success:
        raise RuntimeError(
            "failed to find parameter estimates: \n{0}".format(result.message)
        )
    return dict(zip(("alpha", "beta"), np.exp(result.x)))
